@@ -18,9 +18,9 @@ import (
 	"k8s.io/api/node/v1beta1"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/fields"
-	"k8s.io/apimachinery/pkg/labels"
 	"k8s.io/apimachinery/pkg/runtime"
 	"k8s.io/apimachinery/pkg/types"
+	"k8s.io/klog"
 	"k8s.io/kubernetes/pkg/kubelet/cm/cpuset"
 	"k8s.io/utils/pointer"
 
@@ -43,6 +43,7 @@ import (
 	"github.com/openshift-kni/performance-addon-operators/functests/utils/profiles"
 	"github.com/openshift-kni/performance-addon-operators/pkg/controller/performanceprofile/components"
 	"github.com/openshift-kni/performance-addon-operators/pkg/controller/performanceprofile/components/machineconfig"
+	profilecomponent "github.com/openshift-kni/performance-addon-operators/pkg/controller/performanceprofile/components/profile"
 )
 
 const (
@@ -131,7 +132,7 @@ var _ = Describe("[rfe_id:27368][performance]", func() {
 
 		It("[test_id:37127] Node should point to right tuned profile", func() {
 			for _, node := range workerRTNodes {
-				tuned := tunedForNode(&node)
+				tuned := nodes.TunedForNode(&node, RunningOnSingleNode)
 				activeProfile, err := pods.ExecCommandOnPod(tuned, []string{"cat", "/etc/tuned/active_profile"})
 				Expect(err).ToNot(HaveOccurred(), "Error getting the tuned active profile")
 				activeProfileName := string(activeProfile)
@@ -203,7 +204,7 @@ var _ = Describe("[rfe_id:27368][performance]", func() {
 
 		It("[test_id:35363][crit:high][vendor:cnf-qe@redhat.com][level:acceptance] stalld daemon is running on the host", func() {
 			for _, node := range workerRTNodes {
-				tuned := tunedForNode(&node)
+				tuned := nodes.TunedForNode(&node, RunningOnSingleNode)
 				_, err := pods.ExecCommandOnPod(tuned, []string{"pidof", "stalld"})
 				Expect(err).ToNot(HaveOccurred())
 			}
@@ -240,17 +241,26 @@ var _ = Describe("[rfe_id:27368][performance]", func() {
 				ksoftirq_prio, err := strconv.Atoi(match[1])
 				Expect(err).ToNot(HaveOccurred())
 
-				rcuc_pid, err := nodes.ExecCommandOnNode([]string{"pgrep", "-f", "rcuc", "-n"}, &node)
+				if profile.Spec.RealTimeKernel == nil ||
+					profile.Spec.RealTimeKernel.Enabled == nil ||
+					*profile.Spec.RealTimeKernel.Enabled != true {
+					Expect(stalld_prio).To(BeNumerically("<", ksoftirq_prio))
+					testlog.Warning("Skip checking rcu since RT kernel is disabled")
+					return
+				}
+				//rcuc/n : kthreads that are pinned to CPUs & are responsible to execute the callbacks of rcu threads .
+				//rcub/n : are boosting kthreads ,responsible to monitor per-cpu arrays of lists of tasks that were blocked while in an rcu read-side critical sections.
+				rcu_pid, err := nodes.ExecCommandOnNode([]string{"pgrep", "-f", "rcu[c,b]", "-n"}, &node)
 				Expect(err).ToNot(HaveOccurred())
-				Expect(rcuc_pid).ToNot(BeEmpty())
-				sched_tasks, err = nodes.ExecCommandOnNode([]string{"chrt", "-ap", rcuc_pid}, &node)
+				Expect(rcu_pid).ToNot(BeEmpty())
+				sched_tasks, err = nodes.ExecCommandOnNode([]string{"chrt", "-ap", rcu_pid}, &node)
 				Expect(err).ToNot(HaveOccurred())
 				match = re.FindStringSubmatch(sched_tasks)
-				rcuc_prio, err := strconv.Atoi(match[1])
+				rcu_prio, err := strconv.Atoi(match[1])
 				Expect(err).ToNot(HaveOccurred())
 
+				Expect(stalld_prio).To(BeNumerically("<", rcu_prio))
 				Expect(stalld_prio).To(BeNumerically("<", ksoftirq_prio))
-				Expect(stalld_prio).To(BeNumerically("<", rcuc_prio))
 			}
 		})
 
@@ -383,52 +393,100 @@ var _ = Describe("[rfe_id:27368][performance]", func() {
 		})
 	})
 
-	Context("Network device queues adjusted by Tuned", func() {
-		It("[test_id:40308][crit:high][vendor:cnf-qe@redhat.com][level:acceptance] Should be set to the profile's reserved CPUs count ", func() {
-			noDeviceFound := true
-			if profile.Spec.Net != nil {
-				reservedSet, err := cpuset.Parse(string(*profile.Spec.CPU.Reserved))
-				Expect(err).ToNot(HaveOccurred())
-				reserveCPUsCount := reservedSet.Size()
-				if profile.Spec.Net.UserLevelNetworking != nil && *profile.Spec.Net.UserLevelNetworking && len(profile.Spec.Net.Devices) == 0 {
-					By("To all non virtual network devices when no devices are specified under profile.Spec.Net.Devices")
-					for _, node := range workerRTNodes {
+	Context("KubeletConfig experimental annotation", func() {
+		var secondMCP *machineconfigv1.MachineConfigPool
+		var secondProfile *performancev2.PerformanceProfile
+		var newRole = "test-annotation"
 
-						cmdGetPhysicalDevices := []string{"find", "/sys/class/net", "-type", "l", "-not", "-lname", "*virtual*", "-printf", "%f "}
-						By(fmt.Sprintf("getting a list of physical network devices: %v", cmdGetPhysicalDevices))
-						tuned := tunedForNode(&node)
-						phyDevs, err := pods.ExecCommandOnPod(tuned, cmdGetPhysicalDevices)
-						Expect(err).ToNot(HaveOccurred())
+		BeforeEach(func() {
+			newLabel := fmt.Sprintf("%s/%s", testutils.LabelRole, newRole)
 
-						for _, d := range strings.Split(string(phyDevs), " ") {
-							if d == "" {
-								continue
-							}
-							// See if the device 'd' supports querying the channels.
-							_, err := pods.ExecCommandOnPod(tuned, []string{"ethtool", "-l", d})
-							if err == nil {
-								cmdCombinedChannelsCurrent := []string{"bash", "-c",
-									fmt.Sprintf("ethtool -l %s | sed -n '/Current hardware settings:/,/Combined:/{s/^Combined:\\s*//p}'", d)}
+			reserved := performancev2.CPUSet("0")
+			isolated := performancev2.CPUSet("1-3")
 
-								By(fmt.Sprintf("using physical network device %s for testing", d))
-								out, err := pods.ExecCommandOnPod(tuned, cmdCombinedChannelsCurrent)
-								Expect(err).NotTo(HaveOccurred())
-								channelCurrentCombined, err := strconv.Atoi(strings.TrimSpace(string(out)))
-								if err != nil {
-									testlog.Warningf(fmt.Sprintf("unable to retrieve current multi-purpose channels hardware settings for device %s on %s; skipping test: %v",
-										d, node.Name, err))
-								} else {
-									Expect(err).ToNot(HaveOccurred())
-									Expect(channelCurrentCombined).To(Equal(reserveCPUsCount), " Channel current combine count does not match the count of reserverd CPUs")
-								}
-								noDeviceFound = false
-							}
-						}
+			secondProfile = &performancev2.PerformanceProfile{
+				TypeMeta: metav1.TypeMeta{
+					Kind:       "PerformanceProfile",
+					APIVersion: performancev2.GroupVersion.String(),
+				},
+				ObjectMeta: metav1.ObjectMeta{
+					Name: "test-annotation",
+					Annotations: map[string]string{
+						"kubeletconfig.experimental": `{"systemReserved": {"memory": "256Mi"}, "kubeReserved": {"memory": "256Mi"}}`,
+					},
+				},
+				Spec: performancev2.PerformanceProfileSpec{
+					CPU: &performancev2.CPU{
+						Reserved: &reserved,
+						Isolated: &isolated,
+					},
+					NodeSelector: map[string]string{newLabel: ""},
+					RealTimeKernel: &performancev2.RealTimeKernel{
+						Enabled: pointer.BoolPtr(true),
+					},
+					NUMA: &performancev2.NUMA{
+						TopologyPolicy: pointer.StringPtr("restricted"),
+					},
+				},
+			}
+			Expect(testclient.Client.Create(context.TODO(), secondProfile)).ToNot(HaveOccurred())
 
-						if noDeviceFound {
-							Skip(fmt.Sprintf("no network devices supporting querying channels found on node %s; skipping test", node.Name))
-						}
-					}
+			machineConfigSelector := profilecomponent.GetMachineConfigLabel(secondProfile)
+			secondMCP = &machineconfigv1.MachineConfigPool{
+				ObjectMeta: metav1.ObjectMeta{
+					Name: "test-annotation",
+					Labels: map[string]string{
+						machineconfigv1.MachineConfigRoleLabelKey: newRole,
+					},
+				},
+				Spec: machineconfigv1.MachineConfigPoolSpec{
+					MachineConfigSelector: &metav1.LabelSelector{
+						MatchLabels: machineConfigSelector,
+					},
+					NodeSelector: &metav1.LabelSelector{
+						MatchLabels: map[string]string{
+							newLabel: "",
+						},
+					},
+				},
+			}
+
+			Expect(testclient.Client.Create(context.TODO(), secondMCP)).ToNot(HaveOccurred())
+		})
+
+		It("should override system-reserved memory", func() {
+			var kubeletConfig *machineconfigv1.KubeletConfig
+
+			Eventually(func() error {
+				By("Getting that new KubeletConfig")
+				configKey := types.NamespacedName{
+					Name:      components.GetComponentName(secondProfile.Name, components.ComponentNamePrefix),
+					Namespace: metav1.NamespaceNone,
+				}
+				kubeletConfig = &machineconfigv1.KubeletConfig{}
+				if err := testclient.GetWithRetry(context.TODO(), configKey, kubeletConfig); err != nil {
+					klog.Warningf("Failed to get the KubeletConfig %q", configKey.Name)
+					return err
+				}
+
+				return nil
+			}, time.Minute, 5*time.Second).Should(BeNil())
+
+			kubeletConfigString := string(kubeletConfig.Spec.KubeletConfig.Raw)
+			Expect(kubeletConfigString).To(ContainSubstring(`"kubeReserved":{"memory":"256Mi"}`))
+			Expect(kubeletConfigString).To(ContainSubstring(`"systemReserved":{"memory":"256Mi"}`))
+		})
+
+		AfterEach(func() {
+			if secondProfile != nil {
+				if err := testclient.Client.Delete(context.TODO(), secondProfile); err != nil {
+					klog.Warningf("failed to delete the performance profile %q: %v", secondProfile.Name, err)
+				}
+			}
+
+			if secondMCP != nil {
+				if err := testclient.Client.Delete(context.TODO(), secondMCP); err != nil {
+					klog.Warningf("failed to delete the machine config pool %q: %v", secondMCP.Name, err)
 				}
 			}
 		})
@@ -1141,7 +1199,7 @@ func execSysctlOnWorkers(workerNodes []corev1.Node, sysctlMap map[string]string)
 }
 
 // execute sysctl command inside container in a tuned pod
-func validateTunedActiveProfile(nodes []corev1.Node) {
+func validateTunedActiveProfile(wrknodes []corev1.Node) {
 	var err error
 	var out []byte
 	activeProfileName := components.GetComponentName(testutils.PerformanceProfileName, components.ProfileNamePerformance)
@@ -1160,8 +1218,8 @@ func validateTunedActiveProfile(nodes []corev1.Node) {
 		}
 	}
 
-	for _, node := range nodes {
-		tuned := tunedForNode(&node)
+	for _, node := range wrknodes {
+		tuned := nodes.TunedForNode(&node, RunningOnSingleNode)
 		tunedName := tuned.ObjectMeta.Name
 		By(fmt.Sprintf("executing the command cat /etc/tuned/active_profile inside the pod %s", tunedName))
 		Eventually(func() string {
@@ -1171,33 +1229,3 @@ func validateTunedActiveProfile(nodes []corev1.Node) {
 			fmt.Sprintf("active_profile is not set to %s. %v", activeProfileName, err))
 	}
 }
-
-// find tuned pod for appropriate node
-func tunedForNode(node *corev1.Node) *corev1.Pod {
-	listOptions := &client.ListOptions{
-		Namespace:     components.NamespaceNodeTuningOperator,
-		FieldSelector: fields.SelectorFromSet(fields.Set{"spec.nodeName": node.Name}),
-		LabelSelector: labels.SelectorFromSet(labels.Set{"openshift-app": "tuned"}),
-	}
-
-	tunedList := &corev1.PodList{}
-	Eventually(func() bool {
-		if err := testclient.Client.List(context.TODO(), tunedList, listOptions); err != nil {
-			return false
-		}
-
-		if len(tunedList.Items) == 0 {
-			return false
-		}
-		for _, s := range tunedList.Items[0].Status.ContainerStatuses {
-			if s.Ready == false {
-				return false
-			}
-		}
-		return true
-
-	}, cluster.ComputeTestTimeout(testTimeout*time.Second, RunningOnSingleNode), testPollInterval*time.Second).Should(BeTrue(),
-		"there should be one tuned daemon per node")
-
-	return &tunedList.Items[0]
-}
